#!/usr/bin/env python
import singer
import singer.metrics as metrics
from singer import metadata
from singer import Transformer

logger = singer.get_logger().getChild('tap-bigcommerce')


def sync_stream(state, instance):
    stream = instance.stream

    with metrics.record_counter(stream.tap_stream_id) as counter:
        for (stream, record) in instance.sync(state):
            counter.increment()

            try:
                with Transformer() as transformer:
                    record = transformer.transform(
                        record,
                        stream.schema.to_dict(),
                        metadata.to_map(stream.metadata)
                    )
                singer.write_record(stream.tap_stream_id, record)

                if counter.value % 1000 == 0:
                    singer.write_state(state)

            except Exception as e:
                logger.error('Handled exception: {error}'.format(error=str(e)))
                continue

        return counter.value



#!/usr/bin/env python3
import os
import singer

from tap_bigcommerce.streams import STREAMS


def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def discover_streams(client):
    streams = []

    for s in STREAMS.values():
        s = s(client)
        schema = singer.resolve_schema_references(s.load_schema())
        streams.append({
            'stream': s.name,
            'tap_stream_id': s.name,
            'schema': schema,
            'metadata': s.load_metadata()
        })

    return {"streams": streams}



#!/usr/bin/env python3
import sys
import json
import singer
from singer import utils, metadata, Catalog

from tap_bigcommerce.client import BigCommerce
from tap_bigcommerce.discover import discover_streams
from tap_bigcommerce.streams import STREAMS
from tap_bigcommerce.sync import sync_stream

REQUIRED_CONFIG_KEYS = [
    "start_date", "client_id", "access_token", "store_hash"
]

logger = singer.get_logger().getChild('tap-bigcommerce')


def do_discover(client):
    logger.info("Starting discover")
    catalog = discover_streams(client)
    json.dump(catalog, sys.stdout, indent=2)
    logger.info("Finished discover")


def stream_is_selected(mdata):
    return mdata.get((), {}).get('selected', False)


def get_selected_streams(catalog):
    selected_stream_names = []
    for stream in catalog.streams:
        mdata = metadata.to_map(stream.metadata)
        if stream_is_selected(mdata):
            selected_stream_names.append(stream.tap_stream_id)
    return selected_stream_names


def populate_class_schemas(catalog, selected_stream_names):
    for stream in catalog.streams:
        if stream.tap_stream_id in selected_stream_names:
            STREAMS[stream.tap_stream_id].stream = stream


def ensure_credentials_are_authorized(client):
    if client.is_authorized() is False:
        raise Exception("BigCommerce Client not authorized.")


def do_sync(client, catalog, state, start_date):
    ensure_credentials_are_authorized(client)
    selected_stream_names = get_selected_streams(catalog)
    populate_class_schemas(catalog, selected_stream_names)

    if state.get('bookmarks') is None:
        state = {'bookmarks': {}}

    for stream in catalog.streams:
        stream_name = stream.tap_stream_id

        mdata = metadata.to_map(stream.metadata)

        if stream_name not in selected_stream_names:
            logger.info("%s: Skipping - not selected", stream_name)
            continue

        singer.write_schema(
            stream_name,
            stream.schema.to_dict(),
            metadata.get(mdata, (), 'table-key-properties')
        )

        logger.info("%s: Starting sync", stream_name)

        instance = STREAMS[stream_name](client)
        instance.stream = stream
        if instance.replication_method == "INCREMENTAL":
            if state['bookmarks'].get(
                stream.tap_stream_id, {}
            ).get(instance.replication_key) is None:
                state['bookmarks'][stream.tap_stream_id] = {
                    instance.replication_key: start_date
                }

        counter_value = sync_stream(state, instance)

        singer.write_state(state)

        logger.info("%s: Completed sync (%s rows)", stream_name, counter_value)

    logger.info("Finished sync")


@utils.handle_top_exception(logger)
def main():

    # DEBUG
    # logger.setLevel('DEBUG')

    # Parse command line arguments
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    config = args.config

    # create instance of BigCommerce client
    bigcommerce = BigCommerce(
        client_id=config['client_id'],
        access_token=config['access_token'],
        store_hash=config['store_hash']
    )

    # If discover flag was passed, run discovery mode and dump output to stdout
    if args.discover:
        do_discover(bigcommerce)
    # Otherwise run in sync mode
    else:
        if args.catalog:
            catalog = args.catalog
        else:
            catalog = Catalog.from_dict(discover_streams(bigcommerce))

        do_sync(
            client=bigcommerce,
            catalog=catalog,
            state=args.state,
            start_date=config['start_date']
        )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
from functools import wraps

import singer
from datetime import datetime, timedelta
from dateutil.parser import parse
from tap_bigcommerce.utilities import to_utc
from tap_bigcommerce.bigcommerce import Bigcommerce

logger = singer.get_logger().getChild('tap-bigcommerce')


def validate(method):

    @wraps(method)
    def _validate(*args, **kwargs):
        if 'replication_key' in kwargs and \
                kwargs['replication_key'] not in ['date_modified', 'id']:
                raise Exception("Client Error - invalid replication_key")

        if 'bookmark' in kwargs and \
                type(kwargs['bookmark']) is not datetime:
                raise Exception(
                    "Client Error - bookmark must be valid datetime"
                )

        return method(*args, **kwargs)

    return _validate


def parse_date_string_arguments(fields):
    """
    Transforms specified input datetime arguments into a datetime object
    """

    if type(fields) is not list:
        fields = [fields]

    def decorator(method):
        @wraps(method)
        def parse_dt(*args, **kwargs):
            for key, value in kwargs.items():
                if key in fields:
                    if type(value) != str:
                        raise Exception(
                            "parse_date_string_arguments expects string value. \
                            {} provided".format(value)
                        )
                    kwargs[key] = parse(value)
            return method(*args, **kwargs)
        return parse_dt

    return decorator


class Client():

    authorized = False

    def is_authorized(self):
        return self.authorized is True


class BigCommerce(Client):

    def __init__(self, client_id, access_token, store_hash):
        self.client_id = client_id
        self.access_token = access_token
        self.store_hash = store_hash
        self.utcnow = singer.utils.now()
        self._reset_session()

    def _reset_session(self):
        try:
            self.api = Bigcommerce(
                client_id=self.client_id,
                store_hash=self.store_hash,
                access_token=self.access_token
            )
            self.authorized = True
        except Exception as e:
            self.authorized = False
            raise e

    def iterdates(self, start_date):
        for n in range(max(int((self.utcnow - start_date).days), 1)):
            start = start_date + timedelta(n)
            end = start_date + timedelta(n + 1)
            yield start, min(end, self.utcnow)

    @parse_date_string_arguments('bookmark')
    @validate
    def orders(self, replication_key, bookmark):

        for order in self.api.resource('orders', {
                'min_date_modified': bookmark.isoformat(),
                'sort': 'date_modified:asc'
        }):
            yield order

    @parse_date_string_arguments('bookmark')
    @validate
    def products(self, replication_key, bookmark):

        for product in self.api.resource('products', {
                'date_modified:min': bookmark.isoformat(),
                'sort': 'date_modified',
                'direction': 'asc'
        }):
            yield product

    @parse_date_string_arguments('bookmark')
    @validate
    def customers(self, replication_key, bookmark):
        """
        Customers endpoint can't sort by date_modified, so resource
        is queried by day to ensure consistent replication key
        """

        for start, end in self.iterdates(bookmark):
            for customer in self.api.resource('customers', {
                    'min_date_modified': start.isoformat(),
                    'max_date_modified': end.isoformat()
            }):
                yield customer

    def coupons(self):

        for coupon in self.api.resource('coupons'):
            yield coupon



#!/usr/bin/env python
import os
import json
import pytz
from singer import resolve_schema_references


def to_utc(dtime):
    return dtime.replace(tzinfo=pytz.UTC)


def get_abs_path(path, file=None):
    if file is None:
        file = __file__
    return os.path.join(
        os.path.dirname(os.path.realpath(file)), path)


class SchemaLoader():

    def __init__(
            self,
            schema_path=get_abs_path('schemas'),
            shared_schemas_path=get_abs_path('schemas/shared')):

        self.schema_path = schema_path
        self.shared_schemas_path = shared_schemas_path

    def load(self, name):
        """
        Load schema from JSON and resolve shared $ref
        """
        refs = {}
        shared_file_names = [
            f for f in os.listdir(self.shared_schemas_path) if os.path.isfile(
                os.path.join(self.shared_schemas_path, f)
            )
        ]

        for shared_file in shared_file_names:
            with open(
                os.path.join(self.shared_schemas_path, shared_file)
            ) as data_file:
                refs[shared_file] = json.load(data_file)

        schema_file = self.schema_path + "/{}.json".format(name)
        with open(schema_file) as f:
            schema = json.load(f)

        schema = resolve_schema_references(schema, refs)

        return schema



#!/usr/bin/env python
from setuptools import setup

setup(
    name="tap-bigcommerce",
    version="0.1.6",
    description="Sync data from your BigCommerce Store",
    author="Chris Goddard",
    url="https://github.com/chrisgoddard",
    classifiers=["Programming Language :: Python :: 3 :: Only"],
    py_modules=["tap_bigcommerce"],
    install_requires=[
        "singer-python==5.4.1",
        "requests==2.21.0",
        "requests-futures==0.9.9"
    ],
    entry_points="""
    [console_scripts]
    tap-bigcommerce=tap_bigcommerce:main
    """,
    packages=["tap_bigcommerce"],
    package_data={
        'tap_bigcommerce': [
            'tap_bigcommerce/schemas/*.json',
            'tap_bigcommerce/schemas/shared/*.json'
        ]
    },
    include_package_data=True,
)



#!/usr/bin/env python
"""
API wrapper for BigCommerce library.

Existing Python library for BigCommerce only supports version 2 of
the API. Some endpoints and filters are only available in version 3
of the API.

BigCommerce class gives a basic wrapper around the 4 required resources:
orders, products, customers and coupons.

Rate-Limiting

BigCommerce provides rate limiting information in their responses. This
wrapper limits requests by pausing after a request is made for a period
determinded by taking the number of milliseconds left in the window and
dividing by the number of requests left in the window. The idea being
to maintain the maximum constant rate of requests rather than using all
requests up and then waiting until the window resets.

This may need to change however once we introduce asyncronous nested
resource fetching.

Asyncronous Nested Resource Requests

Standard resources request 50 results per request. However, each result
contains a number of nested resources (Orders, for example, has OrderProducts,
OrderCoupons and ShippingAddress resources).

This creates a problem of extremely slow throughput, because each primary
resource result page will generate a mininum of 150 requests - all executed
syncronously, delaying iteration through the primary resource.

To solve this, sub-resources are requested asyncronously using
requests-futures. A page of results is requested syncronously from the API,
then all of the nested resources for that page of results are requested at
once. Then, in a final loop, the nested resource result is resolved.

In testing, this created a 10-fold increase in speed.

API Rate Limit:

In order to accomodate rate limit restrictions for asyncrounous requests,
at the start of a resource request, the number of requests that will be
needed for the entire page or results is calculated. That number is then
compared to the `X-Rate-Limit-Requests-Left` header. If not enough requests
are available, the script will wait until the next window.

For BigCommerce enterprise plans, the API rate limit is very higher than
the possible throughput of this script. However for low cost plans, the
rate limit is very low (150 requests per 30 seconds), which means that many
resources couldn't be extracted within a single window. To accomodate this,
the initial authorization request checks the rate limit window quota, and
compares this to the number of requests needed. If more requests are required
than are available, the results_per_page number is lowered to a point that will
allow an entire page of results to be processed within one window.

"""

import time
import math

from concurrent.futures import Future
from requests_futures.sessions import FuturesSession
from requests.exceptions import HTTPError
from singer.utils import strptime_to_utc, strftime
from singer import get_logger


logger = get_logger().getChild('tap-bigcommerce')


def filter_excluded_paths(obj, exclude_paths=[]):
    """
    Recurisvely traverse an object and remove fields
    if they match a tuple path (parent, child) provided
    in the list of exclude_paths
    """
    def _filter(o, parent_key=()):
        if type(o) == dict:
            obj = {}
            for key, value in o.items():
                path = parent_key + (key,)
                if path not in exclude_paths:
                    obj[key] = _filter(value, path)
            return obj
        elif type(o) == list:
            return [_filter(el, parent_key) for el in o]
        else:
            return o

    return _filter(obj)


def transform_dates(obj, date_fields=[]):
    """
    Transform dates if the field key is in the provided
    list of fields `date_fields`
    """
    def _transform(o):
        if type(o) == dict:
            obj = {}
            for key, value in o.items():
                if (key in date_fields) and (value not in (None, "")):
                    try:
                        value = strftime(strptime_to_utc(value))
                    except Exception as e:
                        pass
                obj[key] = _transform(value)
            return obj
        elif type(o) == list:
            return [_transform(el) for el in o]
        else:
            return o

    return _transform(obj)


def unpack_nested_resources(get, exclude_fields=[], asyncronous=True):
    """
    Returns a function that will recursively "unpack" an object
    for nested resources by making an asyncrounous request (if
    asyncronous is True). Value of the field will be a Future.
    """

    def unpack(row, parent_key=()):
        if type(row) == dict:
            obj = {}
            for key, value in row.items():
                path = parent_key + (key,)
                if path not in exclude_fields:
                    if type(value) == dict and 'resource' in value:
                        value = get(value['url'], {})
                        if asyncronous is False:
                            value = value.result().data
                    obj[key] = unpack(value, path)
            return obj
        elif type(row) == list:
            return [unpack(el, parent_key) for el in row]
        else:
            return row

    return unpack


def resolve_resources(row, parent_key=()):
    """
    Recurisvely traverse object and Resolve any field values
    that are Futures.
    """
    if type(row) == Future:
        r = row.result()
        return r.data
    if type(row) == dict:
        obj = {}
        for key, value in row.items():
            path = parent_key + (key,)
            obj[key] = resolve_resources(value, path)
        return obj
    elif type(row) == list:
        return [resolve_resources(el, parent_key) for el in row]
    else:
        return row


class Bigcommerce():

    auth_check_url = "https://api.bigcommerce.com/store"

    base_url = "https://api.bigcommerce.com/stores/"

    results_per_page = 50

    max_retries = 5

    retry_window = 300

    """
    Mapping to standardize between the two BigCommerce API
    versions.
    """
    endpoints = {
        'orders': {
            'version': 2,
            'path': 'orders',
            'transform_date_fields': [
                'date_modified',
                'date_created',
                'date_shipped'
            ],
            'sub_resources': 3,
            # deprecated or non-functioning fields
            'exclude_paths': [
                ('credit_card_type',),
                ('products', 'configurable_fields'),
                ('products', 'fulfillment_source'),
                ('shipping_addresses', 'shipping_quotes')
            ]
        },
        'customers': {
            'version': 2,
            'path': 'customers',
            'transform_date_fields': [
                'date_modified',
                'date_created'
            ],
            'sub_resources': 0,
            'exclude_paths': [
                ('addresses',)
            ]
        },
        'products': {
            'version': 3,
            'path': 'catalog/products',
        },
        'coupons': {
            'version': 2,
            'path': 'coupons',
            'transform_date_fields': [
                'date_created',
                'expires'
            ]
        }
    }

    rate_limit = {
        "ms_until_reset": None,
        "window_size_ms": None,
        "requests_remaining": None,
        "requests_quota": None
    }

    def __init__(self, client_id, access_token, store_hash):

        self.retries = 0
        self.last_retry = None
        self.client_id = client_id
        self.access_token = access_token
        self.store_hash = store_hash

        self.base_url = self.base_url + self.store_hash + '/v{version}'

        self._reset_session()

    def _reset_session(self):
        """
        Sets the self.session object to a new FutureSession
        instance and sets default headers and responce hook.

        Called when class instantiated as well as if there is
        an API error and the session needs to be reset.
        """
        self.request_count = 0
        self.session = FuturesSession()

        self.session.hooks['response'] = self._response_hook

        self.headers = {
            'accept': "application/json",
            'content-type': "application/json",
            'x-auth-client': self.client_id,
            'x-auth-token': self.access_token
        }

        # auth check and get rate limit window
        r = self.get(self.make_url(2, 'time'), resolve=True)
        if r.status_code:
            logger.info("BigCommerce API Authorized.")

    def _response_hook(self, resp, *args, **kwargs):
        self.request_count += 1
        if 'X-Rate-Limit-Time-Reset-Ms' in resp.headers:
            self.rate_limit = self._update_rate_limit(resp.headers)

        if resp.status_code != 200:
            if resp.status_code == 204:
                resp.data = []
            else:
                raise HTTPError(resp)
        else:
            resp.data = resp.json()

    def _update_rate_limit(self, headers):
        """
        Parse header object and return a clean dictionary
        of integer values.
        """
        ref = {
            'ms_until_reset': 'X-Rate-Limit-Time-Reset-Ms',
            'window_size_ms': 'X-Rate-Limit-Time-Window-Ms',
            'requests_remaining': 'X-Rate-Limit-Requests-Left',
            'requests_quota': 'X-Rate-Limit-Requests-Quota'
        }
        rate_limit = {}
        for key, header in ref.items():
            rate_limit[key] = int(headers[header])
        return rate_limit

    def make_url(self, version=2, *res):
        """
        Make a valid URL based on the API version and paths provided
        """
        url = self.base_url.format(version=version)
        for r in res:
            url = '{}/{}'.format(url, r)
        return url

    def get(self, url, params={}, resolve=False):
        """
        Make a get request.

        Args:
            url (str): full URL to request
            params (dict): dict of key values for request parameters
            resolve (bool): if True, resolve future before returning
                            (making method blocking), otherwise
                            return Future

        Returns:
            requests.Response
            OR
            concurrent.futures.Future
        """
        future = self.session.get(url, params=params, headers=self.headers)

        if resolve:
            return future.result()
        else:
            return future

    def resource(self, name, params={}, async_sub_resources=True):
        resource = self.endpoints.get(name, {})
        version = resource.get('version', 3)
        path = resource.get('path', name)
        date_fields = resource.get('transform_date_fields', [])
        exclude_paths = resource.get('exclude_paths', [])
        url = self.make_url(version, path)

        unpack_resources = unpack_nested_resources(
            self.get,
            exclude_paths,
            async_sub_resources
        )

        sub_resources = resource.get('sub_resources', 0)

        # adjust results per page based on number of sub resources and
        # the request quota set by the initial authorization check request
        if sub_resources > 0:
            self.results_per_page = min(
                self.results_per_page,
                math.floor(
                    self.rate_limit['requests_quota'] / sub_resources
                ) - 5
            )

        requests_need = self.results_per_page * sub_resources

        page = 0
        while True:
            error_count = 0
            page += 1

            params = {**params, **{
                'page': page,
                'limit': self.results_per_page
            }}

            r = self.get(url, params).result()

            if self.rate_limit['requests_remaining'] is not None:
                if (self.rate_limit['requests_remaining'] - requests_need) < 1:
                    sec = self.rate_limit['ms_until_reset'] / 1000
                    logger.warning(
                        "Rate limit exhausted. Waiting {:.2f} sec".format(sec)
                    )
                    self.request_count = 0
                    time.sleep(sec)

            data = r.data if version == 2 else r.data.get('data', [])
            # unpack nested resources for entire page of results
            data = unpack_resources(data)

            try:
                for row in data:
                    yield transform_dates(
                        filter_excluded_paths(
                            resolve_resources(row),
                            exclude_paths),
                        date_fields)
            except Exception as e:
                error_count += 1
                logger.warning(
                    "Error {} occurred. Sleeping for 10 seconds.".format(e)
                )
                time.sleep(10)
                # max 4 errors in any single page of results
                if error_count > 3:
                    logger.error("{} errors, ending".format(error_count))
                    raise e

            # assume results page with fewer values than `results_per_page` =
            # no more results
            if len(data) < self.results_per_page:
                break



#!/usr/bin/env python
from singer import metadata
from singer import utils
from datetime import timedelta
import os
import singer
import tap_bigcommerce.utilities as tap_utils


logger = singer.get_logger().getChild('tap-bigcommerce')

schema_loader = tap_utils.SchemaLoader()


def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


class Stream():
    name = None
    replication_method = 'INCREMENTAL'
    replication_key = 'date_modified'
    stream = None
    key_properties = ['id']
    bookmark_start = None
    session_bookmark = None
    filter_records_by_bookmark = False
    auto_select_fields = True
    sync_full_table_every = 24

    def __init__(self, client):
        self.client = client

    def get_bookmark(self, state):
        return singer.get_bookmark(state, self.name, self.replication_key)

    def is_bookmark_old(self, value, bookmark):
        if value is None:
            return False
        if bookmark is None:
            return True

        if self.replication_key in ['date_modified', 'date_created']:
            return utils.strptime_with_tz(
                value) > utils.strptime_with_tz(bookmark)
        else:
            return value > bookmark

    def update_session_bookmark_if_old(self, value):
        if self.session_bookmark is None:
            self.session_bookmark = value

        if self.is_bookmark_old(value, self.session_bookmark):
            self.session_bookmark = value

    def update_bookmark_if_old(self, state):
        if self.is_bookmark_old(
            self.session_bookmark,
            self.get_bookmark(state)
        ):
            singer.write_bookmark(
                state,
                self.name,
                self.replication_key,
                self.session_bookmark
            )

    def load_schema(self):
        return schema_loader.load(self.name)

    def load_field_metadata(self, mdata, schema, parent=()):
        if 'object' in schema.get('type', []):
            for field_name, field_schema in schema['properties'].items():
                inclusion = 'automatic' if (
                    field_name in self.key_properties or
                    field_name == self.replication_key
                ) and (
                    parent == ()
                ) else 'available'

                breadcrumb = parent + ('properties', field_name)

                mdata = metadata.write(
                    mdata,
                    breadcrumb,
                    'inclusion',
                    inclusion
                )

                mdata = self.load_field_metadata(
                    mdata, field_schema, breadcrumb)

        elif 'array' in schema.get('type', []):
            mdata = self.load_field_metadata(
                mdata, schema.get('items', {}), parent + ('items',))

        return mdata

    def load_metadata(self):
        schema = self.load_schema()

        mdata = metadata.new()

        mdata = metadata.write(
            mdata, (), 'table-key-properties', self.key_properties
        )

        mdata = metadata.write(
            mdata, (), 'forced-replication-method', self.replication_method
        )

        mdata = metadata.write(
            mdata, (), 'selected-by-default', True
        )

        if self.replication_key:
            mdata = metadata.write(
                mdata, (), 'valid-replication-keys', [self.replication_key]
            )

        mdata = self.load_field_metadata(mdata, schema)

        return metadata.to_list(mdata)

    def is_selected(self):
        return self.stream is not None

    # The main sync function.
    def sync(self, state):
        get_data = getattr(self.client, self.name)

        if self.replication_method == "INCREMENTAL":
            self.bookmark_start = self.get_bookmark(state)
            res = get_data(
                replication_key=self.replication_key,
                bookmark=self.bookmark_start
            )
            for i, item in enumerate(res):
                try:
                    replication_value = item[self.replication_key]

                    if self.is_bookmark_old(replication_value,
                                            self.bookmark_start):

                        yield (self.stream, item)

                        self.update_session_bookmark_if_old(replication_value)
                        self.update_bookmark_if_old(state)

                except Exception as e:
                    logger.error(
                        'Handled exception: {error}'.format(error=str(e))
                    )
                    pass

        elif self.replication_method == "FULL_TABLE":
            res = get_data()

            for item in res:
                yield (self.stream, item)

        else:
            raise Exception(
                'Replication method not defined for {stream}'.format(
                    stream=self.name
                )
            )


class Orders(Stream):
    name = "orders"


class Products(Stream):
    name = "products"


class Coupons(Stream):
    name = "coupons"
    replication_method = "FULL_TABLE"
    replication_key = None


class Customers(Stream):
    name = "customers"


STREAMS = {
    'products': Products,
    'coupons': Coupons,
    'customers': Customers,
    'orders': Orders
}
